"""-----------------------------------------------------------------------------

initialize.py (Last Updated: 10/16/2020)

The purpose of this script is to initialize the rt-cloud session. Specifically,
it will initiate any variables that need to be initiated (e.g., configuration
file) and upload any necessary files to the cloud.

-----------------------------------------------------------------------------"""

# print a short introduction on the internet window
print(""
    "-----------------------------------------------------------------------------\n"
    "The initialize script has not been implemented for this project.\n\n"
    "To implement, please edit initialize.py. For an example as to how you can\n"
    "do this, refer to the initialize.py script in the sample project for the\n"
    "rtcloud framework. Go to https://github.com/brainiak/rt-cloud.git for the\n"
    "repo and navigate to projects/sample/\n"
    "-----------------------------------------------------------------------------")