"""-----------------------------------------------------------------------------

finalize.py (Last Updated: 10/16/2020)

The purpose of this script is to finalize the rt-cloud session. Specifically,
here we want to dowload any important files from the cloud back to the stimulus
computer and maybe even delete files from the cloud that we don't want to 
stay there (maybe for privacy purposes).

-----------------------------------------------------------------------------"""

print(""
    "-----------------------------------------------------------------------------\n"
    "The finalize script has not been implemented for this project.\n\n"
    "To implement, please edit finalize.py. For an example as to how you can\n"
    "do this, refer to the finalize.py script in the sample project for the\n"
    "rtcloud framework. Go to https://github.com/brainiak/rt-cloud.git for the\n"
    "repo and navigate to projects/sample/\n"
    "-----------------------------------------------------------------------------")